/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

__constant__ float4  c_all[ MAXNPROJECTIONS  ] ;	//!< (cosinuses, sinuses, axis position, min bin
texture<float, 2, cudaReadModeElementType> tex_projes;

__global__ static void  hst_cuda_pack_kernel(cufftReal *out, int dpitch, cufftReal *data, int spitch) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    int pos = 2 * idy * spitch + idx;

    tmp1 = data[pos];
    tmp2 = data[pos + spitch];

    pos = 2 * (idy * dpitch + idx);

    out[pos] = tmp1;
    out[pos + 1] = tmp2;
}

__global__ static void  hst_cuda_pack_kernel_zpad(cufftReal *out, int dpitch, cufftReal *data, int spitch, int num_bins) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    int pos;

    if (idx < num_bins) {
        pos = 2 * idy * spitch + idx;

        tmp1 = data[pos];
        tmp2 = data[pos + spitch];
    } else {
        tmp1 = 0;
        tmp2 = 0;
    }

    pos = 2 * (idy * dpitch + idx);

    out[pos] = tmp1;
    out[pos + 1] = tmp2;
}

__global__ static void  hst_cuda_pack_kernel_epad(cufftReal *out, int dpitch, cufftReal *data, int spitch, int num_bins, int mid) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;
    int pos;

    if (idx < num_bins) {
        pos = 2 * idy * spitch + idx;

        tmp1 = data[pos];
        tmp2 = data[pos + spitch];
    } else if (idx < mid) {
        pos = 2 * idy * spitch + num_bins - 1;

        tmp1 = data[pos];
        tmp2 = data[pos + spitch];
    } else {
        pos = 2 * idy * spitch;

        tmp1 = data[pos];
        tmp2 = data[pos + spitch];
    }

    pos = 2 * (idy * dpitch + idx);

    out[pos] = tmp1;
    out[pos + 1] = tmp2;
}


__global__ static void  hst_cuda_filter_kernel(int vector_size, float *data, const float *filter) {
    int idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int id = idx + (blockIdx.y * BLOCK_SIZE + threadIdx.y)*vector_size;

    data[id] *= filter[idx];
}

/*
 left_side:
    flat_zone_start = (axis_position_corr - flat_zone)
    dx = flat_zone_start - i
    multiplier = (2 * prof_shift + prof_fact)  + dx * (prof_fact / pente_zone)

    (2 * prof_shift + prof_fact) is always 1
    prof_fact is -1 or 1

    1 + prof_fact * min(1, (flat_zone_start - i) / pente_zone)
    prof_fact < 0
        param2 is prof_fact / pente_zone
	param1 is flat_zone_start + 1/param2
	multiplier = max(0, (param1 - i)*param2)
    else
        param2 is prof_fact / pente_zone
	param1 is flat_zone_start + 1/param2
	multiplier = min(2, (param1 - i)*param2)

 right_side:
    flat_zone_start = (axis_position_corr + flat_zone)
    dx = flat_zone_start - i

    multiplier = (2 * prof_shift + prof_fact)  + dx * (prof_fact / pente_zone)

    (2 * prof_shift + prof_fact) is always 1
    prof_fact is -1 or 1

    1 + prof_fact * max(-1, (flat_zone_start - i) / pente_zone)
    prof_fact < 0
        param2 is prof_fact / pente_zone
	param1 is flat_zone_start + 1/param2
	multiplier = min(2, (param1 - i)*param2)
    else
        param2 is prof_fact / pente_zone
	param1 is flat_zone_start + 1/param2
	multiplier = max(0, (param1 - i)*param2)

    __global__ static void  limit_kernel_max(float *data, float param1, float param2) {
	int  i = blockIdx.x * BLOCK_SIZE + threadIdx.x;

	data[i] *= max(0., (param1 - i) * param2);
    }

    __global__ static void  limit_kernel_min(float *data, float param1, float param2) {
	int  i = blockIdx.x * BLOCK_SIZE + threadIdx.x;

	data[i] *= min(2., (param1 - i) * param2);
    }

 the next functions are joint

    __global__ static void  limit_kernel_positive(float *data, int first_right, float param1a, float param1b, float param2) {
	int  i = blockIdx.x * BLOCK_SIZE + threadIdx.x;
	int flag = i < first_right;

	data[i] *= min(2., max(flag?1.:0., ((flag?param1a:param1b) - i) * param2));
    }

    __global__ static void  limit_kernel_negative(float *data, int first_right, float param1a, float param1b, float param2) {
	int  i = blockIdx.x * BLOCK_SIZE + threadIdx.x;
	int flag = i < first_right;

	data[i] *= max(0., min(flag?1.:2., ((flag?param1a:param1b) - i) * param2));
    }
 and now completely blocked and joint:
__global__ static void  hst_cuda_limit_kernel(float *data, int pitch, int ystart, float *params) {
    int idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int idy = blockIdx.y * BLOCK_SIZE + threadIdx.y + ystart;
    int i = idy * pitch + idx;

    float axis_position_corr = c_axiss[idy];
    float flat_zone = params[idy * 2];
    float param = params[idy * 2 + 1];

    int flag = idx < (axis_position_corr + flat_zone);

    float multiplier = ((flag?((axis_position_corr - flat_zone) + param):((axis_position_corr + flat_zone) + param)) - idx)/param;


    if (param > 0) {
	data[i] *= min(2., max(flag?1.:0., multiplier));
    } else {
	data[i] *= max(0., min(flag?1.:2., multiplier));
    }
}

*/

__global__ static void  hst_cuda_unpack_kernel(cufftReal *out, int dpitch, cufftReal *data, int spitch) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    int pos = 2 * (idy * spitch + idx);

    tmp1 = data[pos];
    tmp2 = data[pos + 1];

    pos = 2 * idy * dpitch + idx ;

    out[pos] = tmp1;
    out[pos + dpitch] = tmp2;
}

__global__ static void  hst_cuda_unpack_kernel_fai360(cufftReal *out, int dpitch, cufftReal *data, int spitch, int half, float *params, int batch) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    int dest_vector = (idx >= half)?1:0;
    int dest_pos = 2 * (idx - dest_vector * half);

    int src_pos = 2 * (idy * spitch + dest_pos) + dest_vector;
    int src_vector = dest_vector;//(src_pos >= 0);

    tmp1 = data[src_pos];
    tmp2 = data[src_pos + 2];

    float axis_position_corr = c_all[batch + 2*idy + src_vector].z;
    float flat_zone = params[idy * 4 + 2*src_vector];
    float param = params[idy * 4 + 2*src_vector + 1];

    int pos = (2 * idy + dest_vector)*dpitch + dest_pos;

    float apc_plus_flat = axis_position_corr + flat_zone;
    float apc_param_flat = apc_plus_flat + param - dest_pos;

    int flag1 = dest_pos < apc_plus_flat;
    float multiplier1 = __fdividef((flag1?(apc_param_flat - 2 * flat_zone):apc_param_flat),param);

    int flag2 =(dest_pos + 1) < apc_plus_flat;
    float multiplier2 = __fdividef((flag2?(apc_param_flat - 2 * flat_zone):apc_param_flat) - 1,param);

    if (param > 0) {
        out[pos] = tmp1 * min(2., max(flag1?1.:0., multiplier1));
        out[pos+1] = tmp2 * min(2., max(flag2?1.:0., multiplier2));
    } else {
        out[pos] = tmp1 * max(0., min(flag1?1.:2., multiplier1));
        out[pos+1] = tmp2 * max(0., min(flag2?1.:2., multiplier2));
    }
}

/*
 This is backup variant. Theoretically, the first function should work slightly faster due to the reduced amount
 of memory accesses. However, it hardly could be seen!
__global__ static void  hst_cuda_unpack_kernel_fai360(cufftReal *out, int dpitch, cufftReal *data, int spitch, int half, float *params, int batch) {
    float tmp1, tmp2;

    int  idx = blockIdx.x * BLOCK_SIZE + threadIdx.x;
    int  idy = blockIdx.y * BLOCK_SIZE + threadIdx.y;

    int pos = 2 * (idy * spitch + idx);

    tmp1 = data[pos];
    tmp2 = data[pos + 1];

    pos = 2 * idy * dpitch + idx ;


    float axis_position_corr = c_axiss[batch + 2*idy];
    float flat_zone = params[idy * 4];
    float param = params[idy * 4 + 1];

    int flag = idx < (axis_position_corr + flat_zone);
    float multiplier = ((flag?((axis_position_corr - flat_zone) + param):((axis_position_corr + flat_zone) + param)) - idx)/param;


    if (param > 0) {
	out[pos] = tmp1 * min(2., max(flag?1.:0., multiplier));
    } else {
	out[pos] = tmp1 * max(0., min(flag?1.:2., multiplier));
    }

    axis_position_corr = c_axiss[batch + (++idy)];
    flat_zone = params[idy * 2];
    param = params[idy * 2 + 1];

    flag = idx < (axis_position_corr + flat_zone);
    multiplier = ((flag?((axis_position_corr - flat_zone) + param):((axis_position_corr + flat_zone) + param)) - idx)/param;

    pos += dpitch;
    if (param > 0) {
	out[pos] = tmp2 * min(2., max(flag?1.:0., multiplier));
    } else {
	out[pos] = tmp2 * max(0., min(flag?1.:2., multiplier));
    }

}
*/



__global__ static void hst_cuda_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    const int idx = blockIdx.x * BLOCK_SIZE_X + threadIdx.x;
    const int idy = blockIdx.y * BLOCK_SIZE_Y + threadIdx.y + batch;

    float h;
    float res;

    res=0;

    const float bx = idx + apos_off_x;
    const float by = idy + apos_off_y;

#pragma unroll 8
    for (int proje=0; proje<num_proj; proje++) {
        const float4 all = c_all[proje];
        h = all.z + bx * all.x - by * all.y;

        res=res+tex2D(tex_projes, h, proje + 0.5f) ;
    }

    d_SLICE[ BLOCK_SIZE_X*gridDim.x*idy + idx ] = res;
}

__global__ static void hst_cuda_nearest_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float2 res[2] = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 2 * tidx;
    const int idy = bidy + 2 * tidy;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;


    __shared__  float   cache_subh[2*BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float   cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[2*tidx][tidy] = subh;
	cache_subh[2*tidx+1][tidy] = subh + all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;


#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos] = tex2D(tex_projes, minh + pos, proje + projf);
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    int4 minh = {
		cache_subh[2 * tidx][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx][i] - (y + 1) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y + 1) * cache_sin[i],
	    };

	    res[0].x += cache[i][minh.x];
	    res[0].y += cache[i][minh.y];
	    res[1].x += cache[i][minh.z];
	    res[1].y += cache[i][minh.w];

	}


	__syncthreads();

    }

    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy    ) + idx) = res[0];
    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy + 1) + idx) = res[1];
}


__global__ static void hst_cuda_linear_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 2 * tidx;
    const int idy = bidy + 2 * tidy;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;


    __shared__  float   cache_subh[2*BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float2   cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X + 1];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[2*tidx][tidy] = subh;
	cache_subh[2*tidx+1][tidy] = subh + all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;


#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos].x = tex2D(tex_projes, minh + pos, proje + projf);
	}

	__syncthreads();

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos].y = cache[tidy][pos + 1].x - cache[tidy][pos].x;
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    float2 c;
	    float4 v1, v2;
	    float4 minh = {
		cache_subh[2 * tidx][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx][i] - (y + 1) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y + 1) * cache_sin[i],
	    };

	    c = cache[i][(int)minh.x];
	    v1.x = c.x; v2.x = c.y;
	    c = cache[i][(int)minh.y];
	    v1.y = c.x; v2.y = c.y;
	    c = cache[i][(int)minh.z];
	    v1.z = c.x; v2.z = c.y;
	    c = cache[i][(int)minh.w];
	    v1.w = c.x; v2.w = c.y;
	    
	    res += v1 + (minh - floorf(minh)) * v2;
	}

	__syncthreads();

    }

    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy    ) + idx) = make_float2(res.x, res.y);
    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy + 1) + idx) = make_float2(res.z, res.w);
}



__global__ static void hst_cuda_oversample4_kernel(int num_proj, int num_bins, float *d_SLICE, float apos_off_x, float apos_off_y, int batch) {
    float2 res[2] = {0.f, 0.f, 0.f, 0.f};
    
    const int tidx = threadIdx.x;
    const int tidy = threadIdx.y;

    const int bidx = 2 * blockIdx.x * BLOCK_SIZE_X;
    const int bidy = batch + 2 * blockIdx.y * BLOCK_SIZE_Y;

    const int idx = bidx + 2 * tidx;
    const int idy = bidy + 2 * tidy;

    const float bx = bidx + apos_off_x;
    const float by = bidy + apos_off_y;

    const float x = idx + apos_off_x;
    const float y = idy + apos_off_y;


    __shared__  float   cache_subh[2*BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __shared__  float   cache_sin[BLOCK_SIZE_Y];
    __shared__  float   cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X];
    
//#pragma unroll 8
    float projf = tidy + 0.5;
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE_Y) {
        float4 all = c_all[proje + tidy];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = 4 * (all.z + x * all.x - minh);
	cache_subh[2*tidx][tidy] = subh;
	cache_subh[2*tidx+1][tidy] = subh + 4 * all.x;
	if (tidx == 0) cache_sin[tidy] = 4 * all.y;


#pragma unroll 12
	for (int i = 0; i < 12; i++) {
	    int pos = i * BLOCK_SIZE_X + tidx;
	    cache[tidy][pos] = tex2D(tex_projes, minh + ((float)pos)/4, proje + projf);
	}

	__syncthreads();

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE_Y; i++) {
	    int4 minh = {
		cache_subh[2 * tidx][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y) * cache_sin[i],
		cache_subh[2 * tidx][i] - (y + 1) * cache_sin[i],
		cache_subh[2 * tidx + 1][i] - (y + 1) * cache_sin[i],
	    };

	    res[0].x += cache[i][minh.x];
	    res[0].y += cache[i][minh.y];
	    res[1].x += cache[i][minh.z];
	    res[1].y += cache[i][minh.w];

	}


	__syncthreads();

    }

    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy    ) + idx) = res[0];
    *(float2*)(d_SLICE + BLOCK_SIZE_X * 2 * gridDim.x * (idy + 1) + idx) = res[1];
}
